#ifndef ROOT_TModel
#define ROOT_TModel
 
#ifndef ROOT_TBuffer
#include <TBuffer.h>
#endif
#ifndef ROOT_TError
#include <TError.h>
#endif
#ifndef ROOT_Rtypes
#include <Rtypes.h>
#endif
 
#include <math.h>

#define LOG_LEN_CHUNKS

class TEvolver {
 
protected:
   Double_t T;		        // Annealing "temperature"
   double *FitLog;
   gsl_rng *rng_inst;           // RNG instance handle
   
   // should this be handled in TModel where mutation occurs?
   double *SMax,*SMin;	// Search space bounds

   Double_t (*GetFitness)(const TMap& map);
   
public:
   void TEvolver(Double_t (*GetFitnessFcn)(const TMap& map));
   
   void Step(const TMap& map);
   bool AcceptMutation(const TMap& map);
   
   virtual ~TEvolver();
   void TEvolver::Streamer(TBuffer &buf);   

};

class TTrial{

protected:
public:
   TTrial(Double_t (*GetFitnessFcn)(const TMap& map));

   TModel *I1, *phi1;
   TModel *I2, *phi2;
   TModel *I3;

   Double_t Fitness;

   
};


//----- inlines ----------------------------------------------------------------

inline TEvolver::TEvolver(Double_t (*GetFitnessFcn)(const TMap& map))
{
   FitLog = new Double_t[LOG_LEN_CHUNKS];
   GetFitness=GetFitnessFcn;
   
   // Setup GSL Random Number Generator (RNG) environment
   // and create a default-type generator instance handle
   gsl_rng_env_setup();
   rng_inst = gsl_rng_alloc(gsl_rng_default);

}

inline TEvolver::~TEvolver()
{
   delete [] FitLog;
}

inline TEvolver::Step(const TMap& map,Double_t Tnew=T)
{
   T=Tnew;      //Adjust "temperature" is specified

   do map.Mutate(map);   //Mutate until acceptable solution found
   while !AcceptMutation(map,T);
   
}

inline bool TEvolver::AcceptMutation(const TMap& map)
{
   // Accept a mutation based on a value drawn from
   // the exponential distribution exp(-x/T)
   return gsl_ran_exponential(rng_inst,T)<(*GetFitness)(map);

}

inline void TEvolver::Streamer(TBuffer &buf)
{

}


#endif
